use derive_more::derive::Into;

use crate::day7::second_ordering;

use super::{hand_cmp, CategorizedHand};

#[derive(Debug, PartialEq, Eq, Into)]
pub struct JokerOrdered(pub CategorizedHand);

impl JokerOrdered {
    pub fn new(value: CategorizedHand) -> Self {
        Self(value)
    }
}

impl Ord for JokerOrdered {
    fn cmp(&self, other: &Self) -> std::cmp::Ordering {
        hand_cmp(
            &self.0,
            &other.0,
            &second_ordering::LABEL_ORDERING_WITH_JOKER,
        )
    }
}
impl PartialOrd for JokerOrdered {
    fn partial_cmp(&self, other: &Self) -> Option<std::cmp::Ordering> {
        Some(self.cmp(other))
    }
}

#[cfg(test)]
mod testing {
    use std::cmp::Ordering;

    use crate::day7::{
        categorized_hand::{CategorizedHand, JokerOrdered},
        dealt_hand::DealtHand,
        second_ordering::{self, JOKER},
    };

    #[test]
    fn order_with_jokers() {
        fn assert_case(left: &str, right: &str, expected: Ordering) {
            let (left, right): (CategorizedHand, CategorizedHand) = (
                (DealtHand::just_label(left), JOKER).into(),
                (DealtHand::just_label(right), JOKER).into(),
            );
            let (left, right) = (JokerOrdered::new(left), JokerOrdered::new(right));
            let actual = left.cmp(&right);
            assert_eq!(expected, actual, "Left: {:?}\nRight: {:?}", left, right);
        }

        assert_case("AAAJJ", "JBBBB", Ordering::Greater);
        assert_case("22JAB", "22234", Ordering::Less);
    }

    #[test]
    fn example_sort_with_joker_as_weakest() {
        let mut input: Vec<JokerOrdered> = vec![
            DealtHand::new("32T3K", 765),
            DealtHand::new("T55J5", 684),
            DealtHand::new("KK677", 28),
            DealtHand::new("KTJJT", 220),
            DealtHand::new("QQQJA", 483),
        ]
        .into_iter()
        .map(|to_map| JokerOrdered::new((to_map, second_ordering::JOKER).into()))
        .collect();
        input.sort();

        insta::assert_debug_snapshot!(input);
    }
}
